package meerkat;

import com.google.common.util.concurrent.FutureCallback;
import com.google.protobuf.ByteString;
import meerkat.bulletinboard.AsyncBulletinBoardClient;
import meerkat.crypto.DigitalSignature;
import meerkat.protobuf.BulletinBoardAPI.UnsignedBulletinBoardMessage;
import meerkat.protobuf.VoterRegistry.VoterID;
import meerkat.protobuf.VoterRegistry.VoterInfo;
import meerkat.protobuf.VoterRegistry.VoterRegistryMessage;
import meerkat.registry.AsyncRegistryCallbacks.GetGroupsCallback;
import meerkat.registry.AsyncRegistryCallbacks.GetVoterCallback;
import meerkat.registry.AsyncRegistryCallbacks.HasVotedCallback;
import meerkat.registry.MessageCollectionUtils;
import meerkat.registry.RegistryTags;
import meerkat.util.BulletinBoardUtils;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.security.SignatureException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import static meerkat.util.BulletinBoardUtils.signBulletinBoardMessage;


//  TODO : add ability to use DB of certificates and add logging

/**
 * Created by Vladimir Eliezer Tokarev on 1/8/2016.
 * Gives the ability to synchronously manage voters information
 */
public class AsyncRegistry implements VoterRegistry{

    protected Collection<DigitalSignature> signers;
    protected AsyncBulletinBoardClient bulletinBoardClient ;

    @Override
    public void init(Collection<DigitalSignature> signers, AsyncBulletinBoardClient communicator) {
        this.signers = signers;
        this.bulletinBoardClient  = communicator;
    }


    @Override
    public void addVoter(VoterInfo voterInfo, FutureCallback<Boolean> callback) throws SignatureException {
        UnsignedBulletinBoardMessage basicMessage =
                UnsignedBulletinBoardMessage.newBuilder()
                        .addTag(RegistryTags.ID_TAG + voterInfo.getId().getId())
                        .addTag(RegistryTags.VOTER_ENTRY_TAG)
                        .setData(voterInfo.toByteString())
                        .setTimestamp(BulletinBoardUtils.getCurrentTimestampProto())
                        .build();

        bulletinBoardClient.postMessage(signBulletinBoardMessage(basicMessage, signers), callback);
    }

    @Override
    public void setVoterGroups(VoterRegistryMessage voterRegistryMessage, FutureCallback<Boolean> callback) throws SignatureException, IOException {
        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        ObjectOutputStream oos = new ObjectOutputStream(bos);
        oos.writeObject(voterRegistryMessage.getGroupIDList());

        UnsignedBulletinBoardMessage.Builder basicMessage =
                UnsignedBulletinBoardMessage.newBuilder()
                        .addTag(RegistryTags.ID_TAG + voterRegistryMessage.getVoterID().getId())
                        .addTag(RegistryTags.ADD_TO_GROUP_TAG)
                        .setData(ByteString.copyFrom(bos.toByteArray()))
                        .setTimestamp(BulletinBoardUtils.getCurrentTimestampProto());

        bulletinBoardClient.postMessage(signBulletinBoardMessage(basicMessage.build(), signers), callback);
    }

    @Override
    public void setVoted(VoterID voterId, FutureCallback<Boolean> callback) throws SignatureException {
        UnsignedBulletinBoardMessage basicMessage =
                UnsignedBulletinBoardMessage.newBuilder()
                        .addTag(RegistryTags.ID_TAG + voterId.getId())
                        .addTag(RegistryTags.VOTE_ACTION_TAG)
                        .setTimestamp(BulletinBoardUtils.getCurrentTimestampProto())
                        .build();

        bulletinBoardClient.postMessage(signBulletinBoardMessage(basicMessage, signers), callback);
    }

    @Override
    public void getGroups(VoterID voterID, FutureCallback<List<meerkat.protobuf.VoterRegistry.GroupID>> callback) {
        List<String> GroupsActionsTags = new ArrayList<String>(2) {{
           add(RegistryTags.ADD_TO_GROUP_TAG);
           add(RegistryTags.ID_TAG + voterID.getId());
        }};

        bulletinBoardClient.readMessages(MessageCollectionUtils.generateFiltersFromTags(GroupsActionsTags),
                new GetGroupsCallback(callback));
    }

    @Override
    public void getVoter(VoterID voterID, FutureCallback<VoterInfo> callback) {
        List<String> addVoterTags = new ArrayList<String>(2) {{
            add(RegistryTags.ID_TAG + voterID.getId());
            add(RegistryTags.VOTER_ENTRY_TAG);
        }};

        bulletinBoardClient.readMessages(MessageCollectionUtils.generateFiltersFromTags(addVoterTags),
                new GetVoterCallback(callback));
    }

    @Override
    public void hasVoted(VoterID voterId, FutureCallback<Boolean> callBack) {
        List<String> setVotedTags = new ArrayList<String>(2) {{
            add(RegistryTags.ID_TAG + voterId.getId());
            add(RegistryTags.VOTE_ACTION_TAG);
        }};

        bulletinBoardClient.readMessages(MessageCollectionUtils.generateFiltersFromTags(setVotedTags),
                new HasVotedCallback(callBack));
    }
}
